import importlib
import sys
from pathlib import Path
from typing import Protocol

import pytest
from punq import InvalidForwardReferenceError

from iisi_app_core import (
    ComponentKind,
    ComponentRegistry,
    ContainerBuilder,
    PortKind,
    PortsAndAdaptersViolationError,
    application,
    component_definition_for,
    driven_adapter,
    driven_port,
    driving_adapter,
    driving_port,
    port_definition_for,
)
from iisi_app_core.bootstrap import autoload


def _create_pkg_file(path: Path, content: str = "") -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(content, encoding="utf-8")


def _component(target: type[object]):
    definition = component_definition_for(target)
    assert definition is not None
    return definition


def _registry(*targets: type[object]) -> ComponentRegistry:
    registry = ComponentRegistry()
    for target in targets:
        registry.add(_component(target))
    return registry


def test_driving_port_decorator_sets_metadata_for_protocols() -> None:
    @driving_port
    class UseCasePort(Protocol):
        def execute(self) -> str: ...

    definition = port_definition_for(UseCasePort)

    assert definition is not None
    assert definition.port is UseCasePort
    assert definition.kind is PortKind.DRIVING


def test_driven_port_decorator_sets_metadata_for_classes() -> None:
    @driven_port
    class UserRepository:
        def get(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    definition = port_definition_for(UserRepository)

    assert definition is not None
    assert definition.port is UserRepository
    assert definition.kind is PortKind.DRIVEN


def test_application_binds_declared_driving_port() -> None:
    @driving_port
    class GetUserProfilePort:
        def get_user_profile(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @application(port=GetUserProfilePort)
    class GetUserProfile(GetUserProfilePort):
        def get_user_profile(self, user_id: str) -> dict[str, str]:
            return {"id": user_id}

    definition = _component(GetUserProfile)

    assert definition.kind is ComponentKind.APPLICATION
    assert definition.interface is GetUserProfilePort
    assert definition.port is GetUserProfilePort


def test_driving_adapter_alias_sets_kind() -> None:
    @driving_adapter
    class DrivingAdapter:
        pass

    definition = _component(DrivingAdapter)

    assert definition.interface is DrivingAdapter
    assert definition.kind is ComponentKind.DRIVING_ADAPTER
    assert definition.port is None


def test_application_requires_explicit_driving_port() -> None:
    with pytest.raises(TypeError, match="application\\(\\) requires port=\\.\\.\\."):

        @application
        class UseCase:
            pass


def test_application_rejects_non_driving_port() -> None:
    @driven_port
    class UserRepository:
        def get(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    with pytest.raises(TypeError, match="@driving_port"):

        @application(port=UserRepository)
        class GetUserProfile(UserRepository):
            def get(self, user_id: str) -> dict[str, str]:
                return {"id": user_id}


def test_driven_adapter_rejects_non_driven_port() -> None:
    @driving_port
    class UseCasePort:
        def execute(self) -> str:
            raise NotImplementedError

    with pytest.raises(TypeError, match="@driven_port"):

        @driven_adapter(port=UseCasePort)
        class Adapter(UseCasePort):
            def execute(self) -> str:
                return "ok"


def test_component_registry_collects_loaded_modules_in_deterministic_order(tmp_path: Path) -> None:
    pkg_root = tmp_path / "samplepkg_registry"
    _create_pkg_file(pkg_root / "__init__.py")
    _create_pkg_file(
        pkg_root / "a.py",
        "from iisi_app_core import register\n@register\nclass Alpha:\n    pass\n",
    )
    _create_pkg_file(
        pkg_root / "b.py",
        "from iisi_app_core import register\n@register\nclass Beta:\n    pass\n",
    )

    sys.path.insert(0, str(tmp_path))
    try:
        package = importlib.import_module("samplepkg_registry")
        modules = autoload(package)
    finally:
        sys.path.remove(str(tmp_path))

    registry = ComponentRegistry.from_modules(modules)

    assert [definition.qualname for definition in registry.definitions()] == ["Alpha", "Beta"]


def test_container_builder_registers_declared_ports_and_concrete_types() -> None:
    @driven_port
    class UserRepository:
        def get(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @driven_adapter(port=UserRepository)
    class DynamoUserRepository(UserRepository):
        def get(self, user_id: str) -> dict[str, str]:
            return {"id": user_id, "name": "Ada"}

    @driving_port
    class GetUserProfilePort:
        def get_user_profile(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @application(port=GetUserProfilePort)
    class GetUserProfile(GetUserProfilePort):
        def __init__(self, repo: UserRepository) -> None:
            self.repo = repo

        def get_user_profile(self, user_id: str) -> dict[str, str]:
            return self.repo.get(user_id)

    @driving_adapter
    class GetUserProfileHandler:
        def __init__(self, app: GetUserProfilePort) -> None:
            self.app = app

        def handle(self, user_id: str) -> dict[str, str]:
            return self.app.get_user_profile(user_id)

    container = ContainerBuilder(_registry(DynamoUserRepository, GetUserProfile, GetUserProfileHandler)).build()

    assert isinstance(container.resolve(UserRepository), DynamoUserRepository)
    assert isinstance(container.resolve(DynamoUserRepository), DynamoUserRepository)
    assert isinstance(container.resolve(GetUserProfilePort), GetUserProfile)
    assert isinstance(container.resolve(GetUserProfile), GetUserProfile)
    assert container.resolve(GetUserProfileHandler).handle("123") == {"id": "123", "name": "Ada"}


def test_container_builder_rejects_driving_adapter_injecting_application_implementation() -> None:
    @driven_port
    class UserRepository:
        def get(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @driven_adapter(port=UserRepository)
    class DynamoUserRepository(UserRepository):
        def get(self, user_id: str) -> dict[str, str]:
            return {"id": user_id}

    @driving_port
    class GetUserProfilePort:
        def get_user_profile(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @application(port=GetUserProfilePort)
    class GetUserProfile(GetUserProfilePort):
        def __init__(self, repo: UserRepository) -> None:
            self.repo = repo

        def get_user_profile(self, user_id: str) -> dict[str, str]:
            return self.repo.get(user_id)

    @driving_adapter
    class GetUserProfileHandler:
        def __init__(self, app: GetUserProfile) -> None:
            self.app = app

    with pytest.raises(PortsAndAdaptersViolationError, match="decorated implementation '.*GetUserProfile'"):
        ContainerBuilder(_registry(DynamoUserRepository, GetUserProfile, GetUserProfileHandler)).build()


def test_container_builder_rejects_application_injecting_driven_adapter_implementation() -> None:
    @driven_port
    class UserRepository:
        def get(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @driven_adapter(port=UserRepository)
    class DynamoUserRepository(UserRepository):
        def get(self, user_id: str) -> dict[str, str]:
            return {"id": user_id}

    @driving_port
    class GetUserProfilePort:
        def get_user_profile(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @application(port=GetUserProfilePort)
    class GetUserProfile(GetUserProfilePort):
        def __init__(self, repo: DynamoUserRepository) -> None:
            self.repo = repo

        def get_user_profile(self, user_id: str) -> dict[str, str]:
            return self.repo.get(user_id)

    with pytest.raises(PortsAndAdaptersViolationError, match="decorated implementation '.*DynamoUserRepository'"):
        ContainerBuilder(_registry(DynamoUserRepository, GetUserProfile)).build()


def test_container_builder_rejects_application_injecting_list_of_driven_adapter_implementations() -> None:
    @driven_port
    class UserRepository:
        def get(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @driven_adapter(port=UserRepository)
    class DynamoUserRepository(UserRepository):
        def get(self, user_id: str) -> dict[str, str]:
            return {"id": user_id}

    @driving_port
    class GetUserProfilePort:
        def get_user_profile(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @application(port=GetUserProfilePort)
    class GetUserProfile(GetUserProfilePort):
        def __init__(self, repos: list[DynamoUserRepository]) -> None:
            self.repos = repos

        def get_user_profile(self, user_id: str) -> dict[str, str]:
            return self.repos[0].get(user_id)

    with pytest.raises(PortsAndAdaptersViolationError, match="decorated implementation '.*DynamoUserRepository'"):
        ContainerBuilder(_registry(DynamoUserRepository, GetUserProfile)).build()


def test_container_builder_rejects_application_bound_to_same_driving_port_twice() -> None:
    @driving_port
    class GetUserProfilePort:
        def get_user_profile(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @application(port=GetUserProfilePort)
    class GetUserProfile(GetUserProfilePort):
        def get_user_profile(self, user_id: str) -> dict[str, str]:
            return {"id": user_id}

    @application(port=GetUserProfilePort)
    class GetUserProfileCached(GetUserProfilePort):
        def get_user_profile(self, user_id: str) -> dict[str, str]:
            return {"id": user_id, "cached": "true"}

    with pytest.raises(PortsAndAdaptersViolationError, match="multiple Application implementations"):
        ContainerBuilder(_registry(GetUserProfile, GetUserProfileCached)).build()


def test_container_builder_rejects_driven_adapter_bound_to_same_driven_port_twice() -> None:
    @driven_port
    class UserRepository:
        def get(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @driven_adapter(port=UserRepository)
    class DynamoUserRepository(UserRepository):
        def get(self, user_id: str) -> dict[str, str]:
            return {"id": user_id}

    @driven_adapter(port=UserRepository)
    class CachedUserRepository(UserRepository):
        def get(self, user_id: str) -> dict[str, str]:
            return {"id": user_id, "cached": "true"}

    with pytest.raises(PortsAndAdaptersViolationError, match="multiple Driven adapter implementations"):
        ContainerBuilder(_registry(DynamoUserRepository, CachedUserRepository)).build()


def test_container_builder_rejects_driving_adapter_depends_on_driven_port() -> None:
    @driven_port
    class UserRepository:
        def get(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @driven_adapter(port=UserRepository)
    class DynamoUserRepository(UserRepository):
        def get(self, user_id: str) -> dict[str, str]:
            return {"id": user_id}

    @driving_adapter
    class InvalidHandler:
        def __init__(self, repo: UserRepository) -> None:
            self.repo = repo

    with pytest.raises(PortsAndAdaptersViolationError, match="Driving adapters may depend only on @driving_port ports"):
        ContainerBuilder(_registry(DynamoUserRepository, InvalidHandler)).build()


def test_container_builder_allows_undecorated_manual_dependencies() -> None:
    class Clock:
        def now(self) -> str:
            return "2026-03-07T12:00:00Z"

    @driven_port
    class UserRepository:
        def get(self, user_id: str) -> dict[str, str]:
            raise NotImplementedError

    @driven_adapter(port=UserRepository)
    class DynamoUserRepository(UserRepository):
        def __init__(self, clock: Clock) -> None:
            self.clock = clock

        def get(self, user_id: str) -> dict[str, str]:
            return {"id": user_id, "time": self.clock.now()}

    container = ContainerBuilder(_registry(DynamoUserRepository)).add_instance(Clock, Clock()).build()

    assert container.resolve(UserRepository).get("123") == {"id": "123", "time": "2026-03-07T12:00:00Z"}


def test_container_builder_preserves_punq_forward_reference_error_for_unresolved_annotations() -> None:
    @driving_port
    class UseCasePort:
        def execute(self) -> str:
            raise NotImplementedError

    @application(port=UseCasePort)
    class UseCase(UseCasePort):
        def __init__(self, missing: "MissingType") -> None:  # type: ignore[name-defined]  # noqa: F821
            self.missing = missing

        def execute(self) -> str:
            return "ok"

    with pytest.raises(InvalidForwardReferenceError, match="MissingType"):
        ContainerBuilder(_registry(UseCase)).build()



"""Component metadata and container assembly."""

import inspect
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from enum import Enum
from types import ModuleType, UnionType
from typing import Annotated, Any, Union, get_args, get_origin, get_type_hints

from punq import Container, Scope  # type: ignore[import-untyped]

_COMPONENT_DEFINITION_ATTR = "__iisi_component_definition__"
_PORT_DEFINITION_ATTR = "__iisi_port_definition__"
_MISSING = object()
_VALIDATED_COMPONENT_KINDS = {
    "application",
    "driven_adapter",
    "driving_adapter",
}


class ComponentKind(str, Enum):
    COMPONENT = "component"
    DRIVING_ADAPTER = "driving_adapter"
    DRIVEN_ADAPTER = "driven_adapter"
    APPLICATION = "application"
    SEED = "seed"


class PortKind(str, Enum):
    DRIVING = "driving_port"
    DRIVEN = "driven_port"


class ComponentLifetime(str, Enum):
    TRANSIENT = "transient"
    SINGLETON = "singleton"


@dataclass(slots=True, frozen=True)
class PortDefinition:
    port: type[Any]
    kind: PortKind
    module: str = ""
    qualname: str = ""

    def __post_init__(self) -> None:
        object.__setattr__(self, "module", self.module or self.port.__module__)
        object.__setattr__(self, "qualname", self.qualname or self.port.__qualname__)


@dataclass(slots=True, frozen=True)
class ComponentDefinition:
    implementation: type[Any]
    interface: type[Any]
    port: type[Any] | None = None
    kind: ComponentKind = ComponentKind.COMPONENT
    lifetime: ComponentLifetime = ComponentLifetime.TRANSIENT
    module: str = ""
    qualname: str = ""

    def __post_init__(self) -> None:
        object.__setattr__(self, "module", self.module or self.implementation.__module__)
        object.__setattr__(self, "qualname", self.qualname or self.implementation.__qualname__)


class DuplicateComponentError(ValueError):
    """Raised when the same component is registered with conflicting metadata."""


class PortsAndAdaptersViolationError(ValueError):
    """Raised when component wiring violates the ports-and-adapters constraints."""


def infer_interface(target: type[Any]) -> type[Any]:
    for base in target.__bases__:
        if base is object:
            continue
        if isinstance(base, type) and getattr(base, "_is_protocol", False):
            return base
        return base
    return target


def component_definition_for(target: type[Any]) -> ComponentDefinition | None:
    """Return component metadata for a class."""

    value = getattr(target, _COMPONENT_DEFINITION_ATTR, None)
    if isinstance(value, ComponentDefinition):
        return value
    return None


def port_definition_for(target: type[Any]) -> PortDefinition | None:
    """Return port metadata for a class or protocol."""

    value = getattr(target, _PORT_DEFINITION_ATTR, None)
    if isinstance(value, PortDefinition):
        return value
    return None


def _decorate_port(target: type[Any], *, kind: PortKind) -> type[Any]:
    definition = PortDefinition(port=target, kind=kind)
    existing = port_definition_for(target)
    if existing is not None and existing != definition:
        raise PortsAndAdaptersViolationError(f"Port '{target.__qualname__}' already has port metadata.")
    setattr(target, _PORT_DEFINITION_ATTR, definition)
    return target


def _decorate_component(
    target: type[Any],
    *,
    interface: type[Any] | None,
    port: type[Any] | None,
    kind: ComponentKind,
    lifetime: ComponentLifetime,
) -> type[Any]:
    definition = ComponentDefinition(
        implementation=target,
        interface=interface or infer_interface(target),
        port=port,
        kind=kind,
        lifetime=lifetime,
    )
    existing = component_definition_for(target)
    if existing is not None and existing != definition:
        raise DuplicateComponentError(f"Component '{target.__qualname__}' already has component metadata.")
    setattr(target, _COMPONENT_DEFINITION_ATTR, definition)
    return target


def register(
    cls: type[Any] | None = None,
    *,
    interface: type[Any] | None = None,
    kind: ComponentKind = ComponentKind.COMPONENT,
    lifetime: ComponentLifetime = ComponentLifetime.TRANSIENT,
) -> type[Any] | Callable[[type[Any]], type[Any]]:
    """Attach component metadata to a class."""

    def decorator(target: type[Any]) -> type[Any]:
        return _decorate_component(target, interface=interface, port=None, kind=kind, lifetime=lifetime)

    if cls is None:
        return decorator
    return decorator(cls)


def _kind_alias(kind: ComponentKind) -> Callable[..., type[Any] | Callable[[type[Any]], type[Any]]]:
    def alias(
        cls: type[Any] | None = None,
        *,
        interface: type[Any] | None = None,
        lifetime: ComponentLifetime = ComponentLifetime.TRANSIENT,
    ) -> type[Any] | Callable[[type[Any]], type[Any]]:
        return register(cls, interface=interface, kind=kind, lifetime=lifetime)

    return alias


def _port_alias(kind: PortKind) -> Callable[[type[Any]], type[Any]]:
    def decorator(target: type[Any]) -> type[Any]:
        return _decorate_port(target, kind=kind)

    return decorator


def _implements_contract(target: type[Any], contract: type[Any]) -> bool:
    return any(base is contract for base in inspect.getmro(target)[1:])


def _validated_port(
    target: type[Any],
    *,
    port: type[Any],
    expected_kind: PortKind,
    decorator_name: str,
) -> PortDefinition:
    port_definition = port_definition_for(port)
    if port_definition is None:
        raise TypeError(
            f"{decorator_name}() requires port '{port.__qualname__}' "
            f"to be decorated with @{expected_kind.value}."
        )
    if port_definition.kind is not expected_kind:
        raise TypeError(
            f"{decorator_name}() requires a @{expected_kind.value} port, "
            f"got @{port_definition.kind.value} '{port.__qualname__}'."
        )
    if not _implements_contract(target, port):
        raise TypeError(f"{target.__qualname__} must inherit from declared port '{port.__qualname__}'.")
    return port_definition


def application(
    cls: type[Any] | None = None,
    *,
    port: type[Any] | object = _MISSING,
    lifetime: ComponentLifetime = ComponentLifetime.TRANSIENT,
) -> type[Any] | Callable[[type[Any]], type[Any]]:
    """Attach application metadata to a class with an explicit driving port."""

    if port is _MISSING:
        raise TypeError("application() requires port=...")
    assert isinstance(port, type)

    def decorator(target: type[Any]) -> type[Any]:
        _validated_port(target, port=port, expected_kind=PortKind.DRIVING, decorator_name="application")
        return _decorate_component(
            target,
            interface=port,
            port=port,
            kind=ComponentKind.APPLICATION,
            lifetime=lifetime,
        )

    if cls is None:
        return decorator
    return decorator(cls)


def driven_adapter(
    cls: type[Any] | None = None,
    *,
    port: type[Any] | object = _MISSING,
    lifetime: ComponentLifetime = ComponentLifetime.TRANSIENT,
) -> type[Any] | Callable[[type[Any]], type[Any]]:
    """Attach driven-adapter metadata to a class with an explicit driven port."""

    if port is _MISSING:
        raise TypeError("driven_adapter() requires port=...")
    assert isinstance(port, type)

    def decorator(target: type[Any]) -> type[Any]:
        _validated_port(target, port=port, expected_kind=PortKind.DRIVEN, decorator_name="driven_adapter")
        return _decorate_component(
            target,
            interface=port,
            port=port,
            kind=ComponentKind.DRIVEN_ADAPTER,
            lifetime=lifetime,
        )

    if cls is None:
        return decorator
    return decorator(cls)


driving_port = _port_alias(PortKind.DRIVING)
driven_port = _port_alias(PortKind.DRIVEN)
driving_adapter = _kind_alias(ComponentKind.DRIVING_ADAPTER)
seed = _kind_alias(ComponentKind.SEED)


@dataclass(slots=True)
class ComponentRegistry:
    """Collect component metadata from modules."""

    _definitions: dict[type[Any], ComponentDefinition] = field(default_factory=dict)

    def add(self, definition: ComponentDefinition) -> ComponentDefinition:
        existing = self._definitions.get(definition.implementation)
        if existing is None:
            self._definitions[definition.implementation] = definition
            return definition
        if existing != definition:
            raise DuplicateComponentError(
                f"Component '{definition.module}.{definition.qualname}' was registered with conflicting metadata."
            )
        return existing

    def collect_module(self, module: ModuleType) -> None:
        for _, value in sorted(vars(module).items()):
            if not inspect.isclass(value) or value.__module__ != module.__name__:
                continue
            definition = component_definition_for(value)
            if definition is not None:
                self.add(definition)

    def collect_modules(self, modules: Iterable[ModuleType]) -> None:
        for module in sorted(modules, key=lambda item: item.__name__):
            self.collect_module(module)

    @classmethod
    def from_modules(cls, modules: Iterable[ModuleType]) -> "ComponentRegistry":
        registry = cls()
        registry.collect_modules(modules)
        return registry

    def definitions(self) -> tuple[ComponentDefinition, ...]:
        return tuple(
            sorted(
                self._definitions.values(),
                key=lambda definition: (
                    definition.module,
                    definition.qualname,
                    definition.interface.__module__,
                    definition.interface.__qualname__,
                ),
            )
        )


@dataclass(slots=True, frozen=True)
class _FactoryBinding:
    interface: type[Any]
    factory: Callable[..., Any]
    lifetime: ComponentLifetime


class ContainerBuilder:
    """Build a punq container from component metadata."""

    def __init__(self, registry: ComponentRegistry | None = None) -> None:
        self.registry = registry or ComponentRegistry()
        self._instance_bindings: list[tuple[type[Any], Any]] = []
        self._factory_bindings: list[_FactoryBinding] = []

    def add_modules(self, modules: Iterable[ModuleType]) -> "ContainerBuilder":
        self.registry.collect_modules(modules)
        return self

    def add_instance(self, interface: type[Any], instance: Any) -> "ContainerBuilder":
        self._instance_bindings.append((interface, instance))
        return self

    def add_factory(
        self,
        interface: type[Any],
        factory: Callable[..., Any],
        *,
        lifetime: ComponentLifetime = ComponentLifetime.TRANSIENT,
    ) -> "ContainerBuilder":
        self._factory_bindings.append(_FactoryBinding(interface=interface, factory=factory, lifetime=lifetime))
        return self

    def build(self) -> Container:
        definitions = self.registry.definitions()
        _validate_ports_and_adapters(definitions)
        container = Container()
        for definition in definitions:
            scope = _scope_for(definition.lifetime)
            container.register(definition.interface, definition.implementation, scope=scope)
            if definition.interface is not definition.implementation:
                container.register(definition.implementation, definition.implementation, scope=scope)
        for binding in self._factory_bindings:
            container.register(binding.interface, binding.factory, scope=_scope_for(binding.lifetime))
        for interface, instance in self._instance_bindings:
            container.register(interface, instance=instance)
        container.register(Container, instance=container)
        container.register(ComponentRegistry, instance=self.registry)
        return container


def _validate_ports_and_adapters(definitions: tuple[ComponentDefinition, ...]) -> None:
    _validate_unique_port_binding(definitions, ComponentKind.APPLICATION, PortKind.DRIVING)
    _validate_unique_port_binding(definitions, ComponentKind.DRIVEN_ADAPTER, PortKind.DRIVEN)

    for definition in definitions:
        if definition.kind.value not in _VALIDATED_COMPONENT_KINDS:
            continue
        _validate_component_port(definition)
        _validate_constructor_dependencies(definition)


def _validate_unique_port_binding(
    definitions: tuple[ComponentDefinition, ...],
    component_kind: ComponentKind,
    expected_port_kind: PortKind,
) -> None:
    owners: dict[type[Any], ComponentDefinition] = {}
    for definition in definitions:
        if definition.kind is not component_kind:
            continue
        port = definition.port
        if port is None:
            raise PortsAndAdaptersViolationError(
                f"{_component_kind_label(component_kind)} '{definition.qualname}' must declare an explicit port."
            )
        port_definition = port_definition_for(port)
        if port_definition is None or port_definition.kind is not expected_port_kind:
            raise PortsAndAdaptersViolationError(
                f"{_component_kind_label(component_kind)} '{definition.qualname}' "
                f"must use a @{expected_port_kind.value}."
            )
        existing = owners.get(port)
        if existing is not None:
            raise PortsAndAdaptersViolationError(
                f"{_port_label(port_definition)} already has multiple "
                f"{_component_kind_label(component_kind)} implementations: "
                f"'{existing.qualname}' and '{definition.qualname}'."
            )
        owners[port] = definition


def _validate_component_port(definition: ComponentDefinition) -> None:
    if definition.kind is ComponentKind.APPLICATION:
        _validate_explicit_component_port(definition, PortKind.DRIVING)
        return
    if definition.kind is ComponentKind.DRIVEN_ADAPTER:
        _validate_explicit_component_port(definition, PortKind.DRIVEN)


def _validate_explicit_component_port(definition: ComponentDefinition, expected_kind: PortKind) -> None:
    port = definition.port
    if port is None:
        raise PortsAndAdaptersViolationError(
            f"{_component_label(definition)} must declare an explicit @{expected_kind.value}."
        )
    port_definition = port_definition_for(port)
    if port_definition is None or port_definition.kind is not expected_kind:
        raise PortsAndAdaptersViolationError(
            f"{_component_label(definition)} must declare an explicit @{expected_kind.value}."
        )
    if definition.interface is not port:
        raise PortsAndAdaptersViolationError(
            f"{_component_label(definition)} must register its declared port as interface."
        )
    if not _implements_contract(definition.implementation, port):
        raise PortsAndAdaptersViolationError(
            f"{_component_label(definition)} must inherit from declared port '{port.__qualname__}'."
        )


def _validate_constructor_dependencies(definition: ComponentDefinition) -> None:
    for parameter_name, dependency in _constructor_dependencies(definition.implementation):
        dependency_definition = component_definition_for(dependency)
        if dependency_definition is not None:
            raise PortsAndAdaptersViolationError(
                f"{_component_label(definition)} parameter '{parameter_name}' depends on decorated implementation "
                f"'{dependency.__qualname__}'. Depend on its port instead."
            )

        port_definition = port_definition_for(dependency)
        if port_definition is None:
            continue

        if definition.kind is ComponentKind.DRIVING_ADAPTER and port_definition.kind is not PortKind.DRIVING:
            raise PortsAndAdaptersViolationError(
                f"{_component_label(definition)} parameter '{parameter_name}' "
                f"depends on {_port_label(port_definition)}. "
                "Driving adapters may depend only on @driving_port ports."
            )
        if definition.kind is ComponentKind.APPLICATION and port_definition.kind is not PortKind.DRIVEN:
            raise PortsAndAdaptersViolationError(
                f"{_component_label(definition)} parameter '{parameter_name}' "
                f"depends on {_port_label(port_definition)}. "
                "Applications may depend only on @driven_port ports."
            )


def _constructor_dependencies(implementation: type[Any]) -> tuple[tuple[str, type[Any]], ...]:
    init = implementation.__init__
    if init is object.__init__:
        return ()

    signature = inspect.signature(init)
    type_hints = _safe_type_hints(init)
    dependencies: list[tuple[str, type[Any]]] = []

    for parameter_name, parameter in signature.parameters.items():
        if parameter_name == "self":
            continue
        if parameter.kind in {inspect.Parameter.VAR_KEYWORD, inspect.Parameter.VAR_POSITIONAL}:
            continue

        for dependency in _dependency_types(type_hints.get(parameter_name)):
            dependencies.append((parameter_name, dependency))

    return tuple(dependencies)


def _safe_type_hints(init: Callable[..., Any]) -> dict[str, Any]:
    try:
        return get_type_hints(init, include_extras=True)
    except (AttributeError, NameError, TypeError):
        return dict(getattr(init, "__annotations__", {}))


def _dependency_types(annotation: Any) -> tuple[type[Any], ...]:
    if annotation is None:
        return ()
    if isinstance(annotation, type):
        return (annotation,)

    origin = get_origin(annotation)
    if origin is Annotated:
        args = get_args(annotation)
        if not args:
            return ()
        return _dependency_types(args[0])

    args = get_args(annotation)
    if origin in {Union, UnionType}:
        return tuple(
            dict.fromkeys(
                dependency
                for candidate in args
                if candidate is not type(None)
                for dependency in _dependency_types(candidate)
            )
        )

    if origin is list:
        return tuple(
            dict.fromkeys(
                dependency
                for candidate in args
                if candidate is not Ellipsis
                for dependency in _dependency_types(candidate)
            )
        )

    return ()


def _component_kind_label(kind: ComponentKind) -> str:
    if kind is ComponentKind.APPLICATION:
        return "Application"
    if kind is ComponentKind.DRIVEN_ADAPTER:
        return "Driven adapter"
    if kind is ComponentKind.DRIVING_ADAPTER:
        return "Driving adapter"
    if kind is ComponentKind.SEED:
        return "Seed"
    return "Component"


def _component_label(definition: ComponentDefinition) -> str:
    return f"{_component_kind_label(definition.kind)} '{definition.qualname}'"


def _port_label(definition: PortDefinition) -> str:
    prefix = "@driving_port" if definition.kind is PortKind.DRIVING else "@driven_port"
    return f"{prefix} '{definition.qualname}'"


def _scope_for(lifetime: ComponentLifetime) -> Scope:
    return Scope.singleton if lifetime is ComponentLifetime.SINGLETON else Scope.transient


__all__ = [
    "ComponentDefinition",
    "ComponentKind",
    "ComponentLifetime",
    "ComponentRegistry",
    "ContainerBuilder",
    "DuplicateComponentError",
    "PortDefinition",
    "PortKind",
    "PortsAndAdaptersViolationError",
    "application",
    "component_definition_for",
    "driven_adapter",
    "driven_port",
    "driving_adapter",
    "driving_port",
    "infer_interface",
    "port_definition_for",
    "register",
    "seed",
]



"""Support exports."""

from .authz import (
    ROLE_ORDER,
    DomainValidationError,
    ForbiddenError,
    Principal,
    PrincipalGetter,
    Role,
    at_least,
    authorize,
    ensure_at_least,
    has_role,
    is_admin,
    is_reader,
    require_role,
)

__all__ = [
    "DomainValidationError",
    "ForbiddenError",
    "Principal",
    "PrincipalGetter",
    "ROLE_ORDER",
    "Role",
    "at_least",
    "authorize",
    "ensure_at_least",
    "has_role",
    "is_admin",
    "is_reader",
    "require_role",
]



import importlib
import importlib.util
import sys
from pathlib import Path
from typing import Annotated, Any

import pytest

from iisi_app_core import ComponentLifetime, PortsAndAdaptersViolationError


def _create_pkg_file(path: Path, content: str = "") -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(content, encoding="utf-8")


def test_fastapi_integration_import_explains_missing_optional_dependencies(monkeypatch: pytest.MonkeyPatch) -> None:
    import iisi_app_core.integrations  # noqa: F401

    module_name = "iisi_app_core.integrations.fastapi"
    source_path = Path(__file__).parents[1] / "src" / "iisi_app_core" / "integrations" / "fastapi.py"
    original_import_module = importlib.import_module

    def fake_import_module(name: str, package: str | None = None) -> Any:
        if name in {"fastapi", "mangum"}:
            raise ImportError(f"missing {name}")
        return original_import_module(name, package)

    monkeypatch.delitem(sys.modules, module_name, raising=False)
    monkeypatch.setattr(importlib, "import_module", fake_import_module)

    spec = importlib.util.spec_from_file_location(module_name, source_path)
    assert spec is not None
    assert spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module

    with pytest.raises(ImportError, match='iisi-app-core\\[fastapi\\]'):
        spec.loader.exec_module(module)

    sys.modules.pop(module_name, None)


@pytest.mark.skipif(importlib.util.find_spec("fastapi") is None, reason="fastapi extra not installed")
@pytest.mark.skipif(importlib.util.find_spec("mangum") is None, reason="mangum extra not installed")
def test_fastapi_builder_autoloads_container_and_resolves_route_dependency(tmp_path: Path) -> None:
    from fastapi import APIRouter, FastAPI
    from fastapi.testclient import TestClient

    from iisi_app_core.integrations.fastapi import FastApiMangumBuilder, from_container

    pkg_root = tmp_path / "samplepkg_fastapi"
    _create_pkg_file(pkg_root / "__init__.py")
    _create_pkg_file(
        pkg_root / "ports.py",
        """
from iisi_app_core import driven_port, driving_port


@driven_port
class UserRepository:
    def get(self, user_id: str) -> dict[str, str]:
        raise NotImplementedError


@driving_port
class GetUserProfilePort:
    def get_user_profile(self, user_id: str) -> dict[str, str]:
        raise NotImplementedError
""",
    )
    _create_pkg_file(
        pkg_root / "adapters.py",
        """
from iisi_app_core import driven_adapter
from .ports import UserRepository


@driven_adapter(port=UserRepository)
class DynamoUserRepository(UserRepository):
    def get(self, user_id: str) -> dict[str, str]:
        return {"id": user_id, "name": "Ada"}
""",
    )
    _create_pkg_file(
        pkg_root / "application.py",
        """
from iisi_app_core import application
from .ports import GetUserProfilePort, UserRepository


@application(port=GetUserProfilePort)
class GetUserProfile(GetUserProfilePort):
    def __init__(self, repo: UserRepository) -> None:
        self.repo = repo

    def get_user_profile(self, user_id: str) -> dict[str, str]:
        return self.repo.get(user_id)
""",
    )

    sys.path.insert(0, str(tmp_path))
    try:
        package = importlib.import_module("samplepkg_fastapi")
        ports_module = importlib.import_module("samplepkg_fastapi.ports")

        def create_app(_) -> FastAPI:
            router = APIRouter()

            @router.get("/users/{user_id}")
            def get_user_profile(
                user_id: str,
                use_case: Annotated[
                    ports_module.GetUserProfilePort,
                    from_container(ports_module.GetUserProfilePort),
                ],
            ) -> dict[str, str]:
                return use_case.get_user_profile(user_id)

            app = FastAPI()
            app.include_router(router)
            return app

        built = FastApiMangumBuilder().autoload(package).app(create_app).build()
        client = TestClient(built.app)
        response = client.get("/users/123")
    finally:
        sys.path.remove(str(tmp_path))

    assert response.status_code == 200
    assert response.json() == {"id": "123", "name": "Ada"}
    assert built.handler is None
    assert built.app.state.container is built.container
    assert built.app.state.registry is built.registry
    assert [module.__name__ for module in built.modules] == [
        "samplepkg_fastapi.adapters",
        "samplepkg_fastapi.application",
        "samplepkg_fastapi.ports",
    ]


@pytest.mark.skipif(importlib.util.find_spec("fastapi") is None, reason="fastapi extra not installed")
@pytest.mark.skipif(importlib.util.find_spec("mangum") is None, reason="mangum extra not installed")
def test_fastapi_builder_passes_manual_bindings_through_to_routes(tmp_path: Path) -> None:
    from fastapi import APIRouter, FastAPI
    from fastapi.testclient import TestClient

    from iisi_app_core.integrations.fastapi import FastApiMangumBuilder, from_container

    pkg_root = tmp_path / "samplepkg_fastapi_manuals"
    _create_pkg_file(pkg_root / "__init__.py")
    _create_pkg_file(
        pkg_root / "ports.py",
        """
from iisi_app_core import driven_port, driving_port


@driven_port
class UserRepository:
    def get(self, user_id: str) -> dict[str, str]:
        raise NotImplementedError


@driving_port
class GetUserProfilePort:
    def get_user_profile(self, user_id: str) -> dict[str, str]:
        raise NotImplementedError
""",
    )
    _create_pkg_file(
        pkg_root / "adapters.py",
        """
from iisi_app_core import driven_adapter
from .ports import UserRepository


class Settings:
    def __init__(self, prefix: str) -> None:
        self.prefix = prefix


class Clock:
    def now(self) -> str:
        return "2026-03-07T12:00:00Z"


@driven_adapter(port=UserRepository)
class DynamoUserRepository(UserRepository):
    def __init__(self, settings: Settings, clock: Clock) -> None:
        self.settings = settings
        self.clock = clock

    def get(self, user_id: str) -> dict[str, str]:
        return {
            "id": user_id,
            "name": f"{self.settings.prefix}-{user_id}",
            "time": self.clock.now(),
        }
""",
    )
    _create_pkg_file(
        pkg_root / "application.py",
        """
from iisi_app_core import application
from .ports import GetUserProfilePort, UserRepository


@application(port=GetUserProfilePort)
class GetUserProfile(GetUserProfilePort):
    def __init__(self, repo: UserRepository) -> None:
        self.repo = repo

    def get_user_profile(self, user_id: str) -> dict[str, str]:
        return self.repo.get(user_id)
""",
    )

    sys.path.insert(0, str(tmp_path))
    try:
        package = importlib.import_module("samplepkg_fastapi_manuals")
        ports_module = importlib.import_module("samplepkg_fastapi_manuals.ports")
        adapters_module = importlib.import_module("samplepkg_fastapi_manuals.adapters")

        def create_app(_) -> FastAPI:
            router = APIRouter()

            @router.get("/users/{user_id}")
            def get_user_profile(
                user_id: str,
                use_case: Annotated[
                    ports_module.GetUserProfilePort,
                    from_container(ports_module.GetUserProfilePort),
                ],
            ) -> dict[str, str]:
                return use_case.get_user_profile(user_id)

            app = FastAPI()
            app.include_router(router)
            return app

        built = (
            FastApiMangumBuilder()
            .autoload(package)
            .add_instance(adapters_module.Settings, adapters_module.Settings("cfg"))
            .add_factory(adapters_module.Clock, adapters_module.Clock, lifetime=ComponentLifetime.SINGLETON)
            .app(create_app)
            .build()
        )
        client = TestClient(built.app)
        response = client.get("/users/123")
    finally:
        sys.path.remove(str(tmp_path))

    assert response.status_code == 200
    assert response.json() == {"id": "123", "name": "cfg-123", "time": "2026-03-07T12:00:00Z"}


@pytest.mark.skipif(importlib.util.find_spec("fastapi") is None, reason="fastapi extra not installed")
@pytest.mark.skipif(importlib.util.find_spec("mangum") is None, reason="mangum extra not installed")
def test_fastapi_builder_can_build_mangum_handler(tmp_path: Path) -> None:
    from fastapi import FastAPI

    from iisi_app_core.integrations.fastapi import FastApiMangumBuilder

    pkg_root = tmp_path / "samplepkg_fastapi_mangum"
    _create_pkg_file(pkg_root / "__init__.py")
    _create_pkg_file(
        pkg_root / "components.py",
        """
from iisi_app_core import driving_port, application


@driving_port
class PingPort:
    def ping(self) -> str:
        raise NotImplementedError


@application(port=PingPort)
class Ping(PingPort):
    def ping(self) -> str:
        return "pong"
""",
    )

    sys.path.insert(0, str(tmp_path))
    try:
        package = importlib.import_module("samplepkg_fastapi_mangum")

        built = (
            FastApiMangumBuilder()
            .autoload(package)
            .app(lambda _: FastAPI())
            .enable_mangum(lifespan="off")
            .build()
        )
    finally:
        sys.path.remove(str(tmp_path))

    assert built.handler is not None
    assert getattr(built.handler, "app") is built.app


@pytest.mark.skipif(importlib.util.find_spec("fastapi") is None, reason="fastapi extra not installed")
@pytest.mark.skipif(importlib.util.find_spec("mangum") is None, reason="mangum extra not installed")
def test_fastapi_builder_preserves_ports_and_adapters_validation(tmp_path: Path) -> None:
    from fastapi import FastAPI

    from iisi_app_core.integrations.fastapi import FastApiMangumBuilder

    pkg_root = tmp_path / "samplepkg_fastapi_invalid"
    _create_pkg_file(pkg_root / "__init__.py")
    _create_pkg_file(
        pkg_root / "components.py",
        """
from iisi_app_core import application, driven_adapter, driven_port, driving_port


@driven_port
class UserRepository:
    def get(self, user_id: str) -> dict[str, str]:
        raise NotImplementedError


@driven_adapter(port=UserRepository)
class DynamoUserRepository(UserRepository):
    def get(self, user_id: str) -> dict[str, str]:
        return {"id": user_id}


@driving_port
class GetUserProfilePort:
    def get_user_profile(self, user_id: str) -> dict[str, str]:
        raise NotImplementedError


@application(port=GetUserProfilePort)
class GetUserProfile(GetUserProfilePort):
    def __init__(self, repo: DynamoUserRepository) -> None:
        self.repo = repo

    def get_user_profile(self, user_id: str) -> dict[str, str]:
        return self.repo.get(user_id)
""",
    )

    sys.path.insert(0, str(tmp_path))
    try:
        package = importlib.import_module("samplepkg_fastapi_invalid")

        with pytest.raises(PortsAndAdaptersViolationError, match="decorated implementation '.*DynamoUserRepository'"):
            FastApiMangumBuilder().autoload(package).app(lambda _: FastAPI()).build()
    finally:
        sys.path.remove(str(tmp_path))



"""Module autoload helpers."""

import importlib
from pathlib import Path
from types import ModuleType
from typing import Any


class ModuleAutoloadError(ImportError):
    """Raised when module import fails during autoload."""

    def __init__(self, module_name: str) -> None:
        super().__init__(f"Could not import module '{module_name}' during autoload.")
        self.module_name = module_name


def _validate_package(package: Any) -> None:
    if not hasattr(package, "__name__") or not hasattr(package, "__path__"):
        raise TypeError("autoload requires an imported package object with __name__ and __path__")


def _iter_modules(package: Any) -> list[str]:
    module_names: list[str] = []
    for base_path in sorted(str(path) for path in package.__path__):
        base = Path(base_path)
        for path in sorted(base.rglob("*.py")):
            if path.name == "__init__.py":
                continue
            relative_parts = path.relative_to(base).with_suffix("").parts
            module_names.append(".".join((package.__name__, *relative_parts)))
    return module_names


def autoload(package: Any) -> tuple[ModuleType, ...]:
    """Import every module under a package."""

    _validate_package(package)
    modules: list[ModuleType] = []
    for module_name in _iter_modules(package):
        try:
            modules.append(importlib.import_module(module_name))
        except Exception as exc:
            raise ModuleAutoloadError(module_name) from exc
    return tuple(modules)


__all__ = ["ModuleAutoloadError", "autoload"]



import iisi_app_core as core


def test_public_api_exports_expected_symbols() -> None:
    assert "autoload" in core.__all__
    assert "ModuleAutoloadError" in core.__all__
    assert "driving_port" in core.__all__
    assert "driven_port" in core.__all__
    assert "ContainerBuilder" in core.__all__
    assert "PortsAndAdaptersViolationError" in core.__all__
    assert "Principal" in core.__all__
    assert "DomainValidationError" in core.__all__
    assert "ForbiddenError" in core.__all__
    assert "build_lambda_handler" not in core.__all__
    assert "EventBridgeEventHandler" not in core.__all__
    assert "feature_enabled" not in core.__all__



import importlib
import sys
from pathlib import Path

import pytest

from iisi_app_core.bootstrap import ModuleAutoloadError, autoload


def _create_pkg_file(path: Path, content: str = "") -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(content, encoding="utf-8")


def test_autoload_imports_all_modules_except_init(tmp_path: Path) -> None:
    pkg_root = tmp_path / "samplepkg_autoload"
    _create_pkg_file(pkg_root / "__init__.py")
    _create_pkg_file(pkg_root / "a.py")
    _create_pkg_file(pkg_root / "sub" / "__init__.py")
    _create_pkg_file(pkg_root / "sub" / "b.py")

    sys.path.insert(0, str(tmp_path))
    try:
        package = importlib.import_module("samplepkg_autoload")
    finally:
        sys.path.remove(str(tmp_path))

    modules = autoload(package)
    again = autoload(package)

    assert [module.__name__ for module in modules] == ["samplepkg_autoload.a", "samplepkg_autoload.sub.b"]
    assert [module.__name__ for module in again] == ["samplepkg_autoload.a", "samplepkg_autoload.sub.b"]


def test_autoload_rejects_non_package_input() -> None:
    with pytest.raises(TypeError):
        autoload(object())


def test_autoload_wraps_module_import_errors(tmp_path: Path) -> None:
    pkg_root = tmp_path / "brokenpkg_autoload"
    _create_pkg_file(pkg_root / "__init__.py")
    _create_pkg_file(pkg_root / "broken.py", "raise RuntimeError('boom')\n")

    sys.path.insert(0, str(tmp_path))
    try:
        package = importlib.import_module("brokenpkg_autoload")
        with pytest.raises(ModuleAutoloadError) as exc:
            autoload(package)
    finally:
        sys.path.remove(str(tmp_path))

    assert exc.value.module_name == "brokenpkg_autoload.broken"



import pytest

from iisi_app_core import (
    DomainValidationError,
    ForbiddenError,
    Principal,
    Role,
    authorize,
    ensure_at_least,
    require_role,
)


def test_principal_of_allows_empty_tenant_id() -> None:
    principal = Principal.of(name="reader@example.com", role=Role.READER, tenant_id="")
    assert principal.tenant_id == ""


def test_principal_of_rejects_non_string_tenant_id() -> None:
    with pytest.raises(DomainValidationError, match="tenant_id must be a string"):
        Principal.of(name="reader@example.com", role=Role.READER, tenant_id=None)  # type: ignore[arg-type]


def test_authorize_allows_higher_role_through_hierarchy() -> None:
    authorize(Principal.of(name="root", role=Role.ADMIN, tenant_id="tenant"), Role.READER)


def test_ensure_at_least_rejects_too_low_role() -> None:
    with pytest.raises(ForbiddenError):
        ensure_at_least(Principal.of(name="guest", role=Role.GUEST, tenant_id="tenant"), Role.READER)


def test_require_role_reads_principal_from_keyword_argument() -> None:
    @require_role(Role.READER)
    def handle(*, principal: Principal) -> str:
        return principal.name

    result = handle(principal=Principal.of(name="alice", role=Role.READER, tenant_id="tenant"))

    assert result == "alice"


def test_require_role_uses_custom_principal_getter() -> None:
    @require_role(Role.ADMIN, principal_getter=lambda args, kwargs: args[0].principal)
    def handle(command) -> str:  # noqa: ANN001, ANN202
        return command.principal.name

    class Command:
        def __init__(self, principal: Principal) -> None:
            self.principal = principal

    result = handle(Command(Principal.of(name="root", role=Role.ADMIN, tenant_id="tenant")))

    assert result == "root"


def test_require_role_raises_type_error_when_principal_is_missing() -> None:
    @require_role(Role.READER)
    def handle() -> str:
        return "ok"

    with pytest.raises(TypeError, match="principal"):
        handle()



"""Authorization helpers."""

from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from functools import wraps
from typing import Any, ParamSpec, TypeVar


class DomainValidationError(ValueError):
    """Raised when a domain value is invalid."""


class ForbiddenError(PermissionError):
    """Raised when the principal is not authorized to perform the action."""


class Role(str, Enum):
    GUEST = "GUEST"
    ADMIN = "ADMIN"
    READER = "READER"


@dataclass(slots=True, frozen=True)
class Principal:
    name: str
    role: Role
    tenant_id: str

    @classmethod
    def of(cls, name: str, role: Role, tenant_id: str) -> "Principal":
        if not isinstance(name, str) or not name.strip():
            raise DomainValidationError("Principal.name must be a non-empty string")
        if not isinstance(tenant_id, str):
            raise DomainValidationError("Principal.tenant_id must be a string")
        return cls(name=name.strip(), role=role, tenant_id=tenant_id.strip())

    @classmethod
    def guest(cls) -> "Principal":
        return cls(name="guest", role=Role.GUEST, tenant_id="guest")

    @classmethod
    def system(cls) -> "Principal":
        return cls(name="system", role=Role.ADMIN, tenant_id="system")

    def to_dict(self) -> dict[str, str]:
        return {
            "name": self.name,
            "role": self.role.value,
            "tenant_id": self.tenant_id,
        }


ROLE_ORDER = {
    Role.GUEST: 0,
    Role.READER: 1,
    Role.ADMIN: 2,
}

P = ParamSpec("P")
R = TypeVar("R")
PrincipalGetter = Callable[[tuple[Any, ...], dict[str, Any]], Principal]


def has_role(subject: Principal, role: Role) -> bool:
    return subject.role is role


def at_least(subject: Principal, minimum_role: Role) -> bool:
    return ROLE_ORDER[subject.role] >= ROLE_ORDER[minimum_role]


def authorize(subject: Principal, required_role: Role, *, use_hierarchy: bool = True) -> None:
    allowed = at_least(subject, required_role) if use_hierarchy else has_role(subject, required_role)
    if not allowed:
        raise ForbiddenError(f"Requires role {required_role.value}")


def ensure_at_least(subject: Principal, minimum_role: Role) -> None:
    authorize(subject, minimum_role, use_hierarchy=True)


def is_admin(subject: Principal) -> bool:
    return has_role(subject, Role.ADMIN)


def is_reader(subject: Principal) -> bool:
    return has_role(subject, Role.READER)


def _resolve_principal(
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    *,
    principal_getter: PrincipalGetter | None,
) -> Principal:
    if principal_getter is not None:
        subject = principal_getter(args, kwargs)
    elif "principal" in kwargs:
        subject = kwargs["principal"]
    else:
        raise TypeError("require_role could not resolve principal; pass principal=... or principal_getter=...")

    if not isinstance(subject, Principal):
        raise TypeError("require_role expected a Principal instance")

    return subject


def require_role(
    required_role: Role,
    *,
    principal_getter: PrincipalGetter | None = None,
    use_hierarchy: bool = True,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Guard a callable with an explicit principal."""

    def decorator(fn: Callable[P, R]) -> Callable[P, R]:
        @wraps(fn)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            authorize(
                _resolve_principal(args, kwargs, principal_getter=principal_getter),
                required_role,
                use_hierarchy=use_hierarchy,
            )
            return fn(*args, **kwargs)

        return wrapper

    return decorator


__all__ = [
    "DomainValidationError",
    "ForbiddenError",
    "Principal",
    "PrincipalGetter",
    "ROLE_ORDER",
    "Role",
    "at_least",
    "authorize",
    "ensure_at_least",
    "has_role",
    "is_admin",
    "is_reader",
    "require_role",
]



"""Opt-in FastAPI and Mangum bootstrap helpers."""

import importlib
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from types import ModuleType
from typing import Any, TypeVar

from punq import Container  # type: ignore[import-untyped]

from ..bootstrap import ComponentLifetime, ComponentRegistry, ContainerBuilder
from ..bootstrap import autoload as autoload_modules


def _load_dependencies() -> tuple[type[Any], Callable[..., Any], type[Any], type[Any]]:
    missing: list[str] = []
    modules: dict[str, Any] = {}

    for module_name in ("fastapi", "mangum"):
        try:
            modules[module_name] = importlib.import_module(module_name)
        except ImportError:
            missing.append(module_name)

    if missing:
        missing_names = ", ".join(sorted(missing))
        raise ImportError(
            "iisi_app_core.integrations.fastapi requires optional dependencies "
            f"({missing_names}). Install them with `pip install \"iisi-app-core[fastapi]\"`."
        ) from None

    fastapi_module = modules["fastapi"]
    mangum_module = modules["mangum"]
    return (
        getattr(fastapi_module, "FastAPI"),
        getattr(fastapi_module, "Depends"),
        getattr(fastapi_module, "Request"),
        getattr(mangum_module, "Mangum"),
    )


FastAPI, Depends, Request, Mangum = _load_dependencies()

T = TypeVar("T")


@dataclass(slots=True, frozen=True)
class BuiltFastApiContext:
    """Resolved bootstrap context shared with the FastAPI app factory."""

    modules: tuple[ModuleType, ...]
    registry: ComponentRegistry
    container: Container


@dataclass(slots=True, frozen=True)
class BuiltFastApiApp(BuiltFastApiContext):
    """Final FastAPI build output."""

    app: FastAPI
    handler: Mangum | None


def get_container(request: Request) -> Container:
    """Read the punq container from FastAPI app state."""

    container = getattr(request.app.state, "container", None)
    if container is None:
        raise RuntimeError("FastAPI app state is missing `container`; build the app with FastApiMangumBuilder.")
    if not isinstance(container, Container):
        raise RuntimeError("FastAPI app state `container` must be a punq.Container instance.")
    return container


def from_container(interface: type[T], *, use_cache: bool = True) -> Any:
    """Create a FastAPI dependency that resolves an interface from the container."""

    def dependency(request: Request) -> T:
        return get_container(request).resolve(interface)

    dependency.__name__ = f"resolve_{interface.__name__}"
    return Depends(dependency, use_cache=use_cache)


def _unique_modules(modules: Iterable[ModuleType]) -> tuple[ModuleType, ...]:
    unique: dict[str, ModuleType] = {}
    for module in modules:
        unique.setdefault(module.__name__, module)
    return tuple(sorted(unique.values(), key=lambda item: item.__name__))


class FastApiMangumBuilder:
    """Build an autoloaded container, FastAPI app, and optional Mangum handler."""

    def __init__(self) -> None:
        self._modules: list[ModuleType] = []
        self._instance_bindings: list[tuple[type[Any], Any]] = []
        self._factory_bindings: list[tuple[type[Any], Callable[..., Any], ComponentLifetime]] = []
        self._app_factory: Callable[[BuiltFastApiContext], FastAPI] | None = None
        self._mangum_kwargs: dict[str, Any] | None = None

    def autoload(self, package: ModuleType) -> "FastApiMangumBuilder":
        self._modules.extend(autoload_modules(package))
        return self

    def add_modules(self, modules: Iterable[ModuleType]) -> "FastApiMangumBuilder":
        self._modules.extend(modules)
        return self

    def add_instance(self, interface: type[Any], instance: Any) -> "FastApiMangumBuilder":
        self._instance_bindings.append((interface, instance))
        return self

    def add_factory(
        self,
        interface: type[Any],
        factory: Callable[..., Any],
        *,
        lifetime: ComponentLifetime = ComponentLifetime.TRANSIENT,
    ) -> "FastApiMangumBuilder":
        self._factory_bindings.append((interface, factory, lifetime))
        return self

    def app(self, factory: Callable[[BuiltFastApiContext], FastAPI]) -> "FastApiMangumBuilder":
        self._app_factory = factory
        return self

    def enable_mangum(self, **kwargs: Any) -> "FastApiMangumBuilder":
        self._mangum_kwargs = dict(kwargs)
        return self

    def build(self) -> BuiltFastApiApp:
        if self._app_factory is None:
            raise ValueError("FastApiMangumBuilder.app(...) must be configured before build().")

        modules = _unique_modules(self._modules)
        registry = ComponentRegistry.from_modules(modules)

        builder = ContainerBuilder(registry)
        for interface, instance in self._instance_bindings:
            builder.add_instance(interface, instance)
        for interface, factory, lifetime in self._factory_bindings:
            builder.add_factory(interface, factory, lifetime=lifetime)

        container = builder.build()
        context = BuiltFastApiContext(modules=modules, registry=registry, container=container)
        app = self._app_factory(context)
        app.state.container = container
        app.state.registry = registry
        app.state.modules = modules

        handler = Mangum(app, **self._mangum_kwargs) if self._mangum_kwargs is not None else None
        return BuiltFastApiApp(
            modules=modules,
            registry=registry,
            container=container,
            app=app,
            handler=handler,
        )


__all__ = [
    "BuiltFastApiApp",
    "BuiltFastApiContext",
    "FastApiMangumBuilder",
    "from_container",
    "get_container",
]



"""Bootstrap exports."""

from .autoload import ModuleAutoloadError, autoload
from .components import (
    ComponentDefinition,
    ComponentKind,
    ComponentLifetime,
    ComponentRegistry,
    ContainerBuilder,
    DuplicateComponentError,
    PortDefinition,
    PortKind,
    PortsAndAdaptersViolationError,
    application,
    component_definition_for,
    driven_adapter,
    driven_port,
    driving_adapter,
    driving_port,
    infer_interface,
    port_definition_for,
    register,
    seed,
)

__all__ = [
    "ComponentDefinition",
    "ComponentKind",
    "ComponentLifetime",
    "ComponentRegistry",
    "ContainerBuilder",
    "DuplicateComponentError",
    "ModuleAutoloadError",
    "PortDefinition",
    "PortKind",
    "PortsAndAdaptersViolationError",
    "application",
    "autoload",
    "component_definition_for",
    "driven_adapter",
    "driven_port",
    "driving_adapter",
    "driving_port",
    "infer_interface",
    "port_definition_for",
    "register",
    "seed",
]
